import scipy as sp
import numpy as np
import scipy.linalg as la
import scipy.linalg.lapack as LP
from numpy import finfo, double, shape, zeros
from LDL2 import LDL_factor, LDL_solve

# ----------------------------------------------------------
# ---------- A = L*L.T -------------------------------------
# ----------------------------------------------------------


def RLL(A,b,mu=5e-15,maxIt=5,tol=1e-4):
    N = len(b)
    A[np.diag_indices_from(A)] += mu
    L, flag = LP.dpotrf(A)
    y, info = LP.dpotrs(L,b)
    sizeY = la.norm(y,2)
    x = y.copy()

    relativeChange, relativeChangeOld, it, delta = 0, 1e15, 0, []
    while it<maxIt:
        y, info = LP.dpotrs(L,y)
        y *= mu
        correctionSize = la.norm(y,2)
        relativeChange = correctionSize/sizeY
        if relativeChange>relativeChangeOld or relativeChange<tol: break
        it += 1
        relativeChangeOld = relativeChange
        delta.append(relativeChange)
        x += y
        
    return x, delta, it


# -----------------------------------------------------------

def RLL1(A,b,mu=5e-15):
    N = len(b)
    A[np.diag_indices_from(A)] += mu
    L, flag = LP.dpotrf(A)
    x0, info = LP.dpotrs(L,b)
    x1, info = LP.dpotrs(L,x0)
    return x0 + mu*x1


# -----------------------------------------------------------

def RLL0(A,b,mu=5e-15):
    N = len(b)
    A[np.diag_indices_from(A)] += mu
    L, flag = LP.dpotrf( A )
    x0, info = LP.dpotrs(L,b)
    return x0


# ---------- RLDL differentiation matrix ----------------

def RLLDM0(A,B,mu):
    N = shape(B)[0]
    D = zeros((N,N))   
    A[np.diag_indices_from(A)] += mu
    L, flag = LP.dpotrf(A)
    
    for k in range(N):
        x0, flag = LP.dpotrs(L, B[:,k])
        D[:,k] = x0
  
    return D.T


# -----------------------------------------------------------


def RLLDM(A,B,mu):
    N = shape(B)[0]
    D = zeros((N,N))
    A[np.diag_indices_from(A)] += mu
    L, flag = LP.dpotrf(A)
    
    for k in range(N):
        x0, flag = LP.dpotrs(L, B[:,k])
        x1, flag = LP.dpotrs(L, x0)
        D[:,k] = x0 + mu*x1
  
    return D.T


# -----------------------------------------------------------
# ---------- A = L*D*L.T ------------------------------------
# -----------------------------------------------------------

def RLDL(A,b,mu=5e-15,maxIt=5,tol=1e-4):
    N = len(b)
    A[np.diag_indices_from(A)] += mu
    L, d = LDL_factor(A)
    y = LDL_solve(L, d, b)
    sizeY = la.norm(y,2)
    x = y.copy()

    relativeChange, relativeChangeOld, it, delta = 0, 1e15, 0, []
    while it<maxIt:
        y = mu*LDL_solve(L, d, y)
        correctionSize = la.norm(y,2)
        relativeChange = correctionSize/sizeY
        if relativeChange>relativeChangeOld or relativeChange<tol: break
        it += 1
        relativeChangeOld = relativeChange
        delta.append(relativeChange)
        x = x + y
        
    return x, delta, it



# -----------------------------------------------------------

def RLDL1(A,b,mu=5e-15):
    N = len(b)
    A[np.diag_indices_from(A)] += mu
    L, d = LDL_factor(A)
    x0 = LDL_solve(L, d, b)
    x1 = mu*LDL_solve(L, d, x0)
    return x0 + x1

# -----------------------------------------------------------

def RLDL0(A,b,mu=5e-15):
    N = len(b)
    A[np.diag_indices_from(A)] += mu
    L, d = LDL_factor(A)
    x0 = LDL_solve(L, d, b)
    return x0


# ---------- RLDL differentiation matrix ----------------


def RLDLDM0(A,B,mu):
    N = shape(B)[0]
    D = zeros((N,N))
    A[np.diag_indices_from(A)] += mu
    L, d = LDL_factor(A)
    
    for k in range(N):
        D[:,k] = LDL_solve(L, d, B[:,k])
  
    return D.T

# -------------------------------------------------


def RLDLDM(A,B,mu):
    N = shape(B)[0]
    D = zeros((N,N))
    A[np.diag_indices_from(A)] += mu
    L, d = LDL_factor(A)
    
    for k in range(N):
        x0 = LDL_solve(L, d, B[:,k])
        x1 = mu*LDL_solve(L, d, x0)
        D[:,k] = x0 + x1
  
    return D.T

# -------------------------------------------------
